const logger = require("./common").createLogger();
const stringify = require("json-stringify");
const randomitem = require("random-item");
const uniquerandom = require("unique-random");
const randomint = require("random-int");
const SPAWN_DETAILS_KEY = "spwanDetails";
const SPAWN_INDEX = "__spawn_index";
const MAX_BOMB_HIT = "maxBombHitAllowed";
const END_REASON = "Game Had Ended";
const GAME_FINISHED = "gameFinished";
const BOMB_HIT_KEY = "bombHitCount";//"_bomb_hit";
const getMeta = require('./common').getMeta;

const spawnNewFruits = function(game) {
  const gameKey = game.getId();
  let spawnDetails = game.getTransient(SPAWN_DETAILS_KEY);
  if (!spawnDetails) return logger.error({key: gameKey},
      'spawn details object not found!!');
  if (spawnDetails.nextSpawnTime > Date.now()) return;
  let currentSpawnTime = spawnDetails.nextSpawnTime;
  logger.trace({key: gameKey}, 'spawning time....');
  //generate a random array of spawn fruit indexes

  let fruitsToBeSpawned = getSpawnList.call(this, game);
  if (logger.trace()) {
    logger.trace({key: gameKey}, 'will spawn fruits %s ',
        stringify(fruitsToBeSpawned));
  }
  sendSpawnOrder.call(this, game, fruitsToBeSpawned);
  let meta = getMeta(game);
  spawnDetails.nextSpawnTime =
    currentSpawnTime + meta.extras.spawnFrequency;
};

const getSpawnList = function(game) {
  const gameKey = game.getId();
  let spawnPattern = game.getTransient(SPAWN_DETAILS_KEY);
  let spawnIndex = game.getTransient(SPAWN_INDEX);
  logger.trace(
      {key: gameKey},
      'Will spawn %s fruits @slot %s',
      spawnPattern.spawnCount,
      spawnIndex,
  );
  let fruitsToBeSpawned = [];
  let specialFruits = spawnPattern.spawnLists[spawnIndex];
  specialFruits.forEach(t => {
    fruitsToBeSpawned.push(t);
  });
  for (let c = 0; c < spawnPattern.spawnCount - specialFruits.length; c++) {
    fruitsToBeSpawned.push(randomitem(this.normalFruitTypes));
  }
  game.addTransient(SPAWN_INDEX, ++spawnIndex);
  game.setGameProp(SPAWN_INDEX, spawnIndex.toString());
  return fruitsToBeSpawned;
};

const endGame = function(game) {
  const gameKey = game.getId();
  let endTime = game.getScheduledStartTime() + game.getDuration();
  if (endTime < Date.now()) {
    //end game
    logger.debug({key: gameKey}, 'Ending fruit ninja instance %s',
        game.getId());
    game.endGame(END_REASON, err => {
      if (err) {
        logger.error({err: err, key: gameKey});
        logger.error({key: gameKey}, 'Error in ending game');
      }
    });
    return true;
  }
  return false;
};
const step = function(game) {
  let gameState = game.getState();
  const gameKey = game.getId();
  switch (gameState) {
    case 20: //running
      if (endGame.call(this, game)) return;
      spawnNewFruits.call(this, game);
      break;
    default:
      logger.trace({key: gameKey}, 'gamestate is %s, nothing to do ',
          gameState);
  }
};

const sendSpawnOrder = function(game, fruitsToBeSpawned) {
  const gameKey = game.getId();
  let ts = Date.now();
  let connectedPlayers = game.getAllPlayerIds();
  let meta = getMeta(game);
  let objectExpiryPeriod = meta.objectExpiry || meta.duration;
  let expiryTime = Date.now() + objectExpiryPeriod;
  let waves = []
  for (let i = 0; i < 3; i++) {
    waves.push(between(1,5));
  }
  let delays = [500,500,500,500,500,500,500,500]
  connectedPlayers.forEach(player => {
    if (JSON.parse(game.getPlayerProp(player, GAME_FINISHED))) return;

    let gameObjects = [];
    let fruitBasketList = [];
    let fruitsSpawned = 0
    for(let w = 0 ; w < waves.length;w++){
      let waveNumber = player + '_' + w + '_' + Date.now();
      for(let j = 0 ; j < waves[w];j++){
        let type = fruitsToBeSpawned[fruitsSpawned++];
        let gameObject = {
          type: type,
          id: player + "_" + type + "_" + (ts + fruitsSpawned),
          owner: player,
          expiryTime: expiryTime
        };
        let propsValue = objectTypes(type);
        gameObject.propsMap = {
          wavenumber : waveNumber,
          pattern : propsValue.pattern,
          direction : propsValue.direction,
          delay : j === waves[w] - 1 ? delays[Math.floor(Math.random()*delays.length)] : 0
        }
        gameObjects.push(gameObject);

        if (type === 'fruitbasket') {
          fruitBasketList = fruitBasketObjectList(gameObject.id, player);
        }
      }
    }
    game.spawnObjects({name: "fruitbasket", objects: fruitBasketList}, false, player);
    game.spawnObjects({name: "fruits" , objects: gameObjects}, false, player);
  });
};

function fruitBasketObjectList(ID, player) {
  let list = ["pineapple","orange","watermelon","lemon"];
  list.splice(Math.floor(Math.random()*list.length), 1);
  let objectList = [];
  for (const object in list) {
    let gameObject = {
      type: list[object],
      id: player + "_" + list[object] + "_" + ID,
      owner: player
    };
    gameObject.propsMap = {
      parentFruitID: ID
    }
    objectList.push(gameObject);
  }
  return objectList;
}

function objectTypes(type) {
  let result = {pattern: 1, direction:0};
  let fruitsType = ['dragonfruit','pomegranate'];
  if (fruitsType.indexOf(type) != -1) {
    result.pattern = between(14,17);
  } else {
    result.pattern = between(1,13);
    result.direction = between(0,1);
  }
  return result;
}

function between(min, max) {
  return Math.floor(
      Math.random() * (max - min + 1) + min
  )
}

const initFruitTypes = function(meta) {
  if (!meta) return;
  //create an object type array
  let fruitTypes = [];
  meta.gameObjects.forEach(obj => {
    if (obj.type === "blade") return;
    fruitTypes.push(obj.type);
  });
  this.fruitTypes = fruitTypes;
  this.normalFruitTypes = meta.extras.normalfruits;
};

const initSpawnPattern = function(meta) {
  let spawnBounds = meta.extras.spwanBounds;
  let totalSpawnEventCount = -1;
  let maxSpawnUpperBound = -1;
  for (let [type, bound] of Object.entries(spawnBounds)) {
    maxSpawnUpperBound =
      bound[1] > maxSpawnUpperBound ? bound[1] : maxSpawnUpperBound;
  }
  if (maxSpawnUpperBound > -1) {
    //we have some bound
    totalSpawnEventCount = maxSpawnUpperBound + 2; //plus 2 for randomization
  }
  let spawnEventCountStatic = meta.duration / meta.extras.spawnFrequency + 2;
  let calculatedSpawnEventCount =
    spawnEventCountStatic > totalSpawnEventCount
      ? spawnEventCountStatic
      : totalSpawnEventCount;
  let spawnCountPerEvent = Math.ceil(
    (spawnEventCountStatic * meta.extras.spawnCount) / calculatedSpawnEventCount
  );
  let calculatedSpawnFreq = Math.ceil(
    meta.duration / calculatedSpawnEventCount
  );
  //lets create the predefine array
  let spawnLists = [];
  for (let i = 0; i < calculatedSpawnEventCount; i++) {
    let spawnList = [];
    spawnLists.push(spawnList);
  }

  for (let [type, bound] of Object.entries(spawnBounds)) {
    let uRandom = uniquerandom(0, calculatedSpawnEventCount - 1);
    for (let c = 0; c < bound[1]; c++) {
      let toAdd = true;
      if (c > bound[0]) {
        toAdd = randomint(0, 1) == 1;
      }
      if (!toAdd) continue;
      let idx = uRandom();
      spawnLists[idx].push(type);
    }
  }
  let timeBounds = meta.extras.spawnTimebounds;
  for (let [type, bounds] of Object.entries(timeBounds)) {
    for (let i = 0; i < bounds.length; i++) {
      let bound = bounds[i];
      let lowTime = bound[0];
      let hiTime = bound[1];
      let lowSlot = Math.floor(lowTime / calculatedSpawnFreq);
      let hiSlot = Math.floor(hiTime / calculatedSpawnFreq);
      let slotIdx = randomint(lowSlot, hiSlot);
      spawnLists[slotIdx].push(type);
    }
  }
  //create spwanDetails Object
  let spawnPattern = {
    spawnFrequency: calculatedSpawnFreq,
    spawnCount: spawnCountPerEvent,
    spawnLists: spawnLists
  };
  if (logger.debug()) {
    logger.debug("Spawn pattern is %s", stringify(spawnPattern));
  }
  return spawnPattern;
};

const startGame = function(game, cb, meta) {
  const gameKey = game.getId();
  let spawnPattern = initSpawnPattern.call(this, meta);
  spawnPattern.nextSpawnTime = Date.now();
  game.addTransient(SPAWN_DETAILS_KEY, spawnPattern);
  game.setGameProp(SPAWN_DETAILS_KEY, JSON.stringify(spawnPattern));
  game.addTransient(SPAWN_INDEX, 0);
  game.setGameProp(SPAWN_INDEX, "0");
  game.setGameProp(MAX_BOMB_HIT, meta.extras.maxBombHitAllowed);
  //get all the players
  let playerIds = game.getAllPlayerIds();
  if (!playerIds) return cb(null, null);
  if (playerIds.length < 2) return cb(null, null);
  //assign a knife to each one of them
  let knives = game.gameObjectByIds("knife1", "knife2");
  game.changeOwner(playerIds[0], "_knife1");
  game.changeOwner(playerIds[1], "_knife2");
  game.setPlayerProp(playerIds[0], GAME_FINISHED, false);
  game.setPlayerProp(playerIds[1], GAME_FINISHED, false);
  game.setPlayerProp(playerIds[0], BOMB_HIT_KEY, 0);
  game.setPlayerProp(playerIds[1], BOMB_HIT_KEY, 0);
  game.save(err => {
    logger.error({err: err, key: gameKey});
    logger.error({key: gameKey}, 'Error in saving gamestate');
  });
  cb(null, null);
};

const resumeGame = function(game) {
  const gameKey = game.getId();
  logger.info({key: gameKey}, 'resuming game ');
  let spawnDetailsString = game.getGameProp(SPAWN_DETAILS_KEY);
  let spawnDetails = null;
  try {
    spawnDetails = JSON.parse(spawnDetailsString);
  } catch (err) {
    logger.error({err: err, key: gameKey});
    logger.error({key: gameKey}, 'Error in parsing spawnPattern String');
  }
  let meta = getMeta(game);
  spawnDetails =
      spawnDetails == null
      ? initSpawnPattern.call(this, meta)
      : spawnDetails;
  game.addTransient(SPAWN_DETAILS_KEY, spawnDetails);
  //
  let spawnIndex = parseInt(game.getGameProp(SPAWN_INDEX));
  spawnIndex = isNaN(spawnIndex) ? 0 : spawnIndex;
  game.addTransient(SPAWN_INDEX, spawnIndex);
  game.setGameProp(MAX_BOMB_HIT, meta.extras.maxBombHitAllowed);
};

const gameLoaded = function(meta) {
  logger.info("loading defs from meta %s", stringify(meta));
  initFruitTypes.call(this, meta);
};

const gameResult = function(game) {
  const gameKey = game.getId();
  //get all playerIds
  if (game.getState() !== 30) {
    return [];
  }
  let playerIds = game.getAllPlayerIds();
  if (logger.debug()) {
    logger.debug({key: gameKey}, 'playerd who played %s', stringify(playerIds));
  }
  if (playerIds == null || playerIds.length !== 2) {
    logger.error({key: gameKey}, 'Fatal error, player count != 2 ');
    throw {err: 'INVALID_PLAYER_COUNT'};
  }
  let p1Score = game.getPlayerScore(playerIds[0]);
  let p2Score = game.getPlayerScore(playerIds[1]);
  logger.info(
      {key: gameKey},
      'scores for players %s=%s, %s=%s',
      playerIds[0],
      p1Score,
      playerIds[1],
      p2Score,
  );
  if (p1Score === p2Score) {
    return [
      { playerId: playerIds[0], winType: 3 },
      { playerId: playerIds[1], winType: 3 }
    ];
  }
  if (p1Score < p2Score) {
    return [
      { playerId: playerIds[0], winType: 2 },
      { playerId: playerIds[1], winType: 1 }
    ];
  }
  return [
    { playerId: playerIds[0], winType: 1 },
    { playerId: playerIds[1], winType: 2 }
  ];
};

module.exports = { startGame, step, gameLoaded, resumeGame, gameResult };
