const logger = require("./common").createLogger();
const stringify = require("json-stringify");
const KNIFE_ID = "_knife";
const BOMB_HIT_KEY = "bombHitCount";//"_bomb_hit";
const GAME_FINISHED = "gameFinished";
const GAME_OVER_SCORE = "Game over after scoring more then opponent"
const COMBO_ARR = "_combo_arr";
const getMeta = require('./common').getMeta;

const playerDisconnection = function(game, playerId) {
  const gameKey = game.getId();
  logger.trace({key: gameKey}, ' player %s  disconnected', playerId);
  delete this.players[playerId];
};

const playerJoined = function(game, playerId) {
  const gameKey = game.getId();
  logger.trace({key: gameKey}, 'player %s joined ', playerId);
  this.players[playerId] = true;
};

const gameObjectsDestroyed = function(game, gameObjectIds, playerId) {
  const gameKey = game.getId();
  if (logger.trace()) {
    logger.trace(
        {key: gameKey},
        'fruitninjascore:player %s destroyed fruits %s',
        playerId,
        stringify(gameObjectIds),
    );
  }
  let meta = getMeta(game);
  let scoreMap = meta.extras.points;
  let totalScore = 0;
  let fruits = game.gameObjectByIds(...gameObjectIds);
  let totalCuts = 0;
  if (logger.trace()) {
    logger.trace({key: gameKey}, 'fruitninjascore:total fruits %s',
        Object.keys(fruits).length);
  }
  let isGameFinished = game.getPlayerProp(playerId, GAME_FINISHED);
  if (JSON.parse(isGameFinished)) return;
  for (let [id, fruit] of Object.entries(fruits)) {
    if (!fruit) continue;
    let score = 0;
    if (fruit.type === "bomb") {
      continue;
    } else {
      score = scoreMap[fruit.type];
    }
    if (!score) continue;
    totalScore += score;
    totalCuts++;
  }
  logger.debug(
      {key: gameKey},
      'fruitninjascore:Total score for player %s is %s, cuts %s ',
      playerId,
      totalScore,
      totalCuts,
  );
  //
  if (totalScore) {
    game.awardScore(playerId, { score: totalScore }, err => {
      if (err) {
        logger.error({err: err, key: gameKey});
        logger.error({key: gameKey}, 'neutrino core error , in awarding score');
      }
    });
    let gameOver = isGameOverAfterScoreUpdate.call(this, game, playerId);
    if (gameOver) {
      game.endGame(GAME_OVER_SCORE, err => {
        if (err) {
          logger.error({err: err, key: gameKey});
          logger.error({key: gameKey}, 'Error in ending game');
        }
      });
    }
    logger.debug(
        {key: gameKey},
        'fruitninjascore:score %s is awarded to player %s ',
        totalScore,
        playerId,
    );
  }

  let combosizeachieved = 0;
  let combopoints = meta.extras.combopoints;
  if (combopoints) {
    let comboPointsMessages = meta.extras.combopointsmessages;
    totalCuts--;
    let combopointsIdx =
      totalCuts > combopoints.length - 1 ? combopoints.length - 1 : totalCuts;
    let comboPointsEarned = combopoints[combopointsIdx];
    logger.trace(
        {key: gameKey},
        'fruitninjascore: combopointsIdx %s, comboPointsEarned %s',
        combopointsIdx,
        comboPointsEarned,
    );
    if (comboPointsEarned > 0) {
      combosizeachieved = combopointsIdx + 1;
      game.awardScore(
          playerId,
          {
            score: comboPointsEarned,
            message: JSON.stringify({
              count: totalCuts + 1,
              displayText: comboPointsMessages[combopointsIdx],
            }),
          },
          err => {
            if (err) {
              logger.error({err: err, key: gameKey});
              logger.error({key: gameKey},
                  'neutrino core error , in awarding score');
            }
          },
      );
      let gameOver = isGameOverAfterScoreUpdate.call(this, game, playerId);
      if (gameOver) {
        game.endGame(GAME_OVER_SCORE, err => {
          if (err) {
            logger.error({err: err, key: gameKey});
            logger.error({key: gameKey}, 'Error in ending game');
          }
        });
      }
      // send game event of size of combo
      game.publishGameEvent(1202, [playerId], {size: combosizeachieved});
    }
  }
  //look for combo of combos
  if (combosizeachieved) {
    checkComboOfCombos.call(this, game, combosizeachieved, playerId);
  }
};

const playerCommunication = function(game){
  let allPlayerIds = game.getAllPlayerIds();
  allPlayerIds.forEach(id => {
    game.syncPlayerServVars(parseInt(id))
  });
}

const checkComboOfCombos = function(game, combosizeachieved, playerId) {
  const gameKey = game.getId();
  let isGameFinished = game.getPlayerProp(playerId, GAME_FINISHED);
  if (JSON.parse(isGameFinished)) return;
  let comboArr = game.getTransient(COMBO_ARR);
  if (!comboArr) {
    comboArr = [];
    game.addTransient(COMBO_ARR, comboArr);
    comboArr.push(combosizeachieved);
    return;
  }
  comboArr.push(combosizeachieved);

  for (let i = 0; i < this.comboofcombossorted; i++) {
    let comboofcombo = this.comboofcombossorted[i];
    if (comboOfComboSatisfied(comboArr, comboofcombo)) {
      //award the additional score
      game.awardScore(
        playerId,
        {
          score: comboofcombo,
          message: comboofcombo.winText
        },
        err => {
          if (err) {
            logger.error({err: err, key: gameKey});
            logger.error({key: gameKey},
                'neutrino core error , in awarding score');
          }
        }
      );
      let gameOver = isGameOverAfterScoreUpdate.call(this, game, playerId);
      if (gameOver) {
        game.endGame(GAME_OVER_SCORE, err => {
          if (err) {
            logger.error({err: err, key: gameKey});
            logger.error({key: gameKey}, 'Error in ending game');
          }
        });
      }
      break;
    }
  }
};

const comboOfComboSatisfied = function(combos, ccdef) {
  if (combos.length < ccdef.minCombosRequired) return false;
  for (let i = combos.length - 1; i >= 0; i--) {
    let combo = combos[i];
    if (combo < ccdef.minSize) return false;
  }
  return true;
};

const isKnife = function(id) {
  return id.startsWith(KNIFE_ID);
};
const collision = function(game, playerId, id1, id2) {
  const gameKey = game.getId();
  let meta = getMeta(game);
  if (!isKnife(id1) && !isKnife(id2)) return;
  let knifeId = null;
  let fruitId = null;
  if (isKnife(id1)) {
    knifeId = id1;
    fruitId = id2;
  } else {
    knifeId = id2;
    fruitId = id1;
  }
  let fruits = game.gameObjectByIds([fruitId]);
  if (!fruits) return;

  let fruit = fruits[fruitId];
  if (!fruit) return;
  let isGameFinished = game.getPlayerProp(playerId, GAME_FINISHED);
  if (JSON.parse(isGameFinished)) return;
  if (fruit.type === "pomegranate") {
    game.awardScore(
        playerId,
        {
          score: 2,
        },
        err => {
          if (err) {
            logger.error({err: err, key: gameKey});
            logger.error({key: gameKey},
                'neutrino core error , in awarding score');
          }
        },
    );
    let gameOver = isGameOverAfterScoreUpdate.call(this, game, playerId);
    if (gameOver) {
      game.endGame(GAME_OVER_SCORE, err => {
        if (err) {
          logger.error({err: err, key: gameKey});
          logger.error({key: gameKey}, 'Error in ending game');
        }
      });
    }
  } else if (fruit.type === 'bomb') {
    let bombhitString = game.getPlayerProp(playerId, BOMB_HIT_KEY);
    let bombhit = parseInt(bombhitString);
    bombhit = isNaN(bombhit) ? 0 : bombhit;
    let score = bombhit > 0 ? -30 : -20;
    game.setPlayerProp(playerId, BOMB_HIT_KEY, (bombhit + 1).toString());
    game.awardScore(
        playerId,
        {score},
        err => {
          if (err) {
            logger.error({err: err, key: gameKey});
            logger.error({key: gameKey},
                'neutrino core error , in awarding score');
          }
        },
    );
    playerCommunication.call(this, game);
    if ((bombhit + 1) >= meta.extras.maxBombHitAllowed) {
      let gameOver = isGameOver.call(this, game, playerId);
      if (!gameOver) {
        game.setPlayerProp(playerId, GAME_FINISHED, true);
        playerCommunication.call(this, game);
      } else {
        game.endGame("Game over after bomb hit", err => {
          if (err) {
            logger.error({err: err, key: gameKey});
            logger.error({key: gameKey}, 'Error in ending game');
          }
        });
      }
    }
    game.publishGameEvent(1201, [playerId]);
  }
};

const getOpponentPlayerId = function(game, playerId) {
  let allPlayerIds = game.getAllPlayerIds();
  let oppId = null;
  allPlayerIds.forEach(id => {
    if (parseInt(id) !== parseInt(playerId)) {
      oppId = id;
    }
  });
  return oppId;
};

const isGameOver = function(game, playerId) {
  const gameKey = game.getId();
  let opponentPlayerId = getOpponentPlayerId.call(this, game, playerId);
  if (JSON.parse(game.getPlayerProp(opponentPlayerId,
      GAME_FINISHED))) return true;
  let playerScore = game.getPlayerScore(playerId);
  let opponentScore = game.getPlayerScore(opponentPlayerId);
  logger.info({key: gameKey}, 'scores of collision player %s and opponent %s', playerScore,
      opponentScore, opponentPlayerId);
  return playerScore < opponentScore;
}

const isGameOverAfterScoreUpdate = function(game, playerId) {
  const gameKey = game.getId();
  let opponentPlayerId = getOpponentPlayerId.call(this, game, playerId);
  if (JSON.parse(
      game.getPlayerProp(opponentPlayerId, GAME_FINISHED))) {
    let playerScore = game.getPlayerScore(playerId)
    let opponentScore = game.getPlayerScore(opponentPlayerId)
    logger.info({key: gameKey},
        'isGameOverAfterScoreUpdate scores of collision player %s and opponent %s',
        playerScore, opponentScore, opponentPlayerId);
    return playerScore > opponentScore;
  }
  return false;
};

module.exports = {
  playerDisconnection,
  playerJoined,
  gameObjectsDestroyed,
  collision
};
